import json
from collections import deque
from typing import Dict, List, Dict, Optional

from autogpt.llm_utils import create_chat_completion
from autogpt.config import Config

response_format = """[
    {
        "task_id": 1,
        "task_description": "task description"
    },
    {
        "task_id": 2,
        "task_description": "task description"
    }
]"""

def create_message(prompt: str) -> Dict[str, str]:
    """Create a message for the chat completion

    Args:
        chunk (str): The chunk of text to summarize
        question (str): The question to answer

    Returns:
        Dict[str, str]: The message to send to the chat completion
    """
    return {
        "role": "user",
        "content": prompt,
    }

def task_creation_agent(objective: str, task: Dict, task_list: List[str], allowed_new_task: Optional[int] = 1):
    prompt = (
        f"You are an task creation AI that uses the result of an execution agent to create new tasks with the following objective: {objective}.\n"
        f"The last completed task has the result: {task['data']}. "
        f"This result was based on this task description: {task['task_description']}. "
        f"These are incomplete tasks: {', '.join(task_list)}. "
        f"Based on the result, create new tasks to be completed by the AI system that do not overlap with incomplete tasks. Up to {allowed_new_task} new tasks can be created.\n"
        "You should only respond in JSON format as described below \n"
        f"Response Format: \n{response_format} \nEnsure the response can be parsed by Python json.loads"
    )

    messages = [create_message(prompt)]
    response = create_chat_completion(
        model=Config().fast_llm_model,
        messages=messages,
    )

    new_tasks = json.loads(response)
    return [{"task_description": task["task_description"]} for task in new_tasks]

def prioritization_agent(this_task_id:int, objective:str, task_list:deque) -> deque:
    task_descriptions = [t["task_description"] for t in task_list]
    next_task_id = int(this_task_id)+1
    prompt = (
        f"You are an task prioritization AI tasked with cleaning the formatting of and reprioritizing the following tasks: {task_descriptions}. "
        f"Consider the ultimate objective of your team: {objective}. \n"
        "Tasks should be sorted from highest to lowest priority.\n"
        "Higher-priority tasks are those that act as pre-requisites or are more essential for meeting the objective.\n"
        f"Do not remove any tasks and don't add any new tasks. Return the result as a numbered list, like:\n"
        "#. First task\n"
        "#. Second task\n"
        f"Start the task list with number {next_task_id}."
    )

    messages = [create_message(prompt)]
    response = create_chat_completion(
        model=Config().fast_llm_model,
        messages=messages,
    )

    new_tasks = response.split('\n')
    task_list = []
    for task_string in new_tasks:
        task_parts = task_string.strip().split(".", 1)
        if len(task_parts) == 2:
            task_id = task_parts[0].strip()
            task_description = task_parts[1].strip()
            task_list.append({"task_id": task_id, "task_description": task_description})
    return task_list

def init_task_list(objective: str, max_tasks_count: Optional[int] = 4) -> List[Dict]:
    task_id_counter = 0
    prompt = (
        f"You are an task creation AI that creates tasks with the following objective: {objective}.\n"
        f"Up to {max_tasks_count - 1} new tasks can be created to complete this objective. "
        "You should only respond in JSON format as described below \n"
        f"Response Format: \n{response_format} \nEnsure the response can be parsed by Python json.loads"
    )
    messages = [create_message(prompt)]
    response = create_chat_completion(
        model=Config().fast_llm_model,
        messages=messages,
    )

    new_tasks = json.loads(response)
    task_list = []
    for new_task in new_tasks:
        task_id_counter += 1
        new_task.update({"task_id": task_id_counter})
        task_list.append(new_task)
    return task_list



from autoxx.agent.base import BaseAgent
from typing import List, Dict
from autoxx.search.simple_search import simple_search
from autogpt.llm_utils import create_chat_completion
from autogpt.config import Config

CFG = Config()

class ResearchAgent(BaseAgent):  
  
    def execute(self, objective:str, task: str) -> str:  
        return simple_search(task)
  
    def post_process(self, objective:str, task_results:List[Dict]) -> str:  
        task_context = [f"Task({task['task_description']}): {task['data']}" for task in task_results]
        nl = '\n'
        prompt = (
            "You are an task processing AI."
            f"Given the context {nl.join(task_context) } and objective: {objective}. "
            "Create a research report. "
        )
        messages = [{
            "role": "user",
            "content": prompt,
        }]
        return create_chat_completion(
            model=CFG.fast_llm_model,
            messages=messages,
        )



import logging, time, threading
from typing import Dict, Optional
from autoxx.agent.base import BaseAgent
from autoxx.agent.researcher import ResearchAgent
from autoxx.agi import init_task_list, task_creation_agent, prioritization_agent

class Task:
    def __init__(self, agent:BaseAgent, objective:str, max_tasks_count: Optional[int] = 3):
        if objective is None or len(objective) <= 1:  
            raise ValueError("Must provide an objective")  
        if agent is None:  
            raise ValueError("Must provide an agent")

        self.agent = agent
        self.objective = objective
        self.max_tasks_count = max_tasks_count

        self.task_list = init_task_list(objective, max_tasks_count)
        self.task_cur = 0
        self.status = "initizing"
        self.error_message = None
        self.final_answer = None

    def execute(self) -> str:
        self.status = "running"
        # Main loop
        while self.task_cur < len(self.task_list):
            # Print the task list
            incomplete_list = [self.task_list[i]  for i in range(self.task_cur, len(self.task_list))]
            logging.info(f"\033[95m\033[1m INCOMPLETE TASK LIST({self.objective}) \033[0m\033[0m {[str(t['task_id'])+': '+t['task_description'] for t in incomplete_list]}")

            # Step 1: Pull the first task
            task = incomplete_list[0]
            logging.info(f"\033[92m\033[1m NEXT TASK({self.objective}) \033[0m\033[0m" + str(task['task_id'])+": "+task['task_description'])

            try:
                # Send to execution function to complete the task based on the context
                result = self.agent.execute(self.objective, task["task_description"])
                this_task_id = int(task["task_id"])
                logging.info(f"\033[93m\033[1m TASK RESULT({self.objective}) \033[0m\033[0m {result}")
            except Exception as e:  
                logging.error(f"An error occurred while executing the task({self.objective}): {e}")
                self.status = "error"
                self.error_message = str(e)  
                return str(e)
                # You can also handle specific exceptions if needed, e.g., KeyError, ValueError, etc.  
                # To do this, add additional except blocks for each specific exception type.  

            task["data"] = result
            self.task_cur += 1

            # Step 2: Create new tasks
            task_id_counter = len(self.task_list)
            if self.max_tasks_count-task_id_counter > 0:
                new_tasks = task_creation_agent(self.objective,  task, [t["task_description"]  for t in self.task_list[self.task_cur:]], self.max_tasks_count-task_id_counter)

                for new_task in new_tasks:
                    task_id_counter += 1
                    new_task.update({"task_id": task_id_counter})
                    self.task_list.append(new_task)

                incomplete_list = incomplete_list[1:]
                # Step 3: Reprioritize task list
                if len(self.task_list)-self.task_cur > 1:
                    reprioritized_tasks = prioritization_agent(this_task_id, self.objective, self.task_list[self.task_cur:])
                    self.task_list = self.task_list[:self.task_cur] + reprioritized_tasks

            if self.task_cur >= self.max_tasks_count:
                final_result = self.agent.post_process(self.objective, self.task_list)
                self.final_answer = final_result
                self.status = "finished"
                return final_result
            time.sleep(1)  # Sleep before checking the task list again

    def stat(self) -> Dict:
        current = self.task_cur
        if current >= len(self.task_list):
            current -= 1
        return {"current_task": self.task_list[current]["task_id"], "task_list": self.task_list, "final_answer": self.final_answer, "status": self.status, "error_message": self.error_message}
  
class TaskManager:  
    def __init__(self):  
        self.tasks = {}  
  
    def create_task(self, objective: str, max_tasks_count: Optional[int] = 3):  
        if objective in self.tasks:
            if self.tasks[objective].status == "error":
                task = self.tasks[objective]
                task_thread = threading.Thread(target=task.execute)  
                task_thread.start()
                return
            else:
                raise ValueError(f"Task found with objective: {objective}")
        
        agent = ResearchAgent()  # Use the default ResearchAgent  
        task = Task(agent, objective, max_tasks_count) 
        self.tasks[objective] = task  
  
        # Start task execution in a separate thread  
        task_thread = threading.Thread(target=task.execute)  
        task_thread.start()
  
    def get_task_stat(self, objective: str) -> Dict:  
        if objective not in self.tasks:  
            raise ValueError(f"No task found with objective: {objective}")  
  
        return self.tasks[objective].stat()


from abc import ABC, abstractmethod  
from typing import List, Dict
  
class BaseAgent(ABC):  
  
    @abstractmethod  
    def execute(self, objective:str, task: str) -> str:  
        pass  
  
    @abstractmethod  
    def post_process(self, objective:str, task_results:List[Dict]) -> str:  
        pass  



from typing import Dict, List, Optional
import json
import os
from pathlib import Path
from autogpt.config import Config
import openai

from autogpt.commands.web_selenium import scrape_text_with_selenium
from selenium.webdriver.remote.webdriver import WebDriver
from autogpt.commands.google_search import google_search
from autogpt import token_counter
from autogpt.llm_utils import create_chat_completion
from autogpt.processing.text import split_text

FILE_DIR = Path(__file__).parent.parent
CFG = Config()

def setup_autgpt_config(
        azure_config_file:Optional[str] = "",
        only_gpt4:bool = False,
        debug:bool = False,
) -> Config:
    CFG = Config()
    if azure_config_file != "":
        CFG.set_openai_api_key(os.getenv("AZURE_OPENAI_API_KEY", ""))
        CFG.use_azure=True
        CFG.load_azure_config(config_file=azure_config_file)

    if only_gpt4:
        CFG.set_fast_llm_model("gpt-4")
        CFG.set_fast_token_limit(8000)
        CFG.set_browse_chunk_max_length(7000)

    CFG.set_debug_mode(debug)

    openai.api_type = CFG.openai_api_type
    openai.api_base = CFG.openai_api_base
    openai.api_version = CFG.openai_api_version
    openai.api_key = CFG.openai_api_key
    return CFG

def create_message(chunk: str, question: str) -> Dict[str, str]:
    """Create a message for the chat completion

    Args:
        chunk (str): The chunk of text to summarize
        question (str): The question to answer

    Returns:
        Dict[str, str]: The message to send to the chat completion
    """
    return {
        "role": "user",
        "content": f'"""{chunk}""" Using the above text, answer the following'
        f' question: "{question}" -- if the question cannot be answered using the text,'
        " summarize the text.",
    }

def scroll_to_percentage(driver: WebDriver, ratio: float) -> None:
    """Scroll to a percentage of the page

    Args:
        driver (WebDriver): The webdriver to use
        ratio (float): The percentage to scroll to

    Raises:
        ValueError: If the ratio is not between 0 and 1
    """
    if ratio < 0 or ratio > 1:
        raise ValueError("Percentage should be between 0 and 1")
    driver.execute_script(f"window.scrollTo(0, document.body.scrollHeight * {ratio});")

def summarize_chunks(text:str, question: str, url: str, driver: Optional[WebDriver] = None) -> List[str]:
    """Summarize a list of chunks

    Args:
        chunks (list[str]): The chunks to summarize
        question (str): The question to answer

    Returns:
        str: The summary of the chunks
    """
    model = CFG.fast_llm_model
    text_length = len(text)
    print(f"Text length: {text_length} characters")

    chunks = list(
        split_text(text=text, max_length=CFG.browse_chunk_max_length, question=question)
    )
    summaries = []

    scroll_ratio = 1 / len(chunks)
    for i, chunk in enumerate(chunks):
        if driver:
            scroll_to_percentage(driver, scroll_ratio * i)

        messages = [create_message(chunk, question)]
        tokens_for_chunk = token_counter.count_message_tokens(messages, model)
        print(
            f"Summarizing {url} chunk {i + 1} / {len(chunks)} of length {len(chunk)} characters, or {tokens_for_chunk} tokens"
        )

        summary = create_chat_completion(
            model=CFG.fast_llm_model,
            messages=messages,
        )
        summaries.append(summary)

    return summaries

def add_header(driver: WebDriver) -> None:
    """Add a header to the website

    Args:
        driver (WebDriver): The webdriver to use to add the header

    Returns:
        None
    """
    driver.execute_script(open(f"{FILE_DIR}/js/overlay.js", "r").read())


def browse_website(url: str, question: str) -> List[str]:
    """Browse a website and summarize it

    Args:
        url (str): The url to browse
        question (str): The question to answer

    Returns:
        str: The summary of the website
    """
    driver, text = scrape_text_with_selenium(url=url)
    add_header(driver)
    summary = summarize_chunks(text=text, question=question, url=url, driver=driver)
    return summary

def simple_search(question:str, search_num:Optional[int]=2) -> str:
    search_result = google_search(query=question, num_results=search_num)
    search_search_json = json.loads(search_result)
    summaries = []

    if isinstance(search_search_json, list):
        for res in search_search_json:
            print(f"{res['title']} x {res['href']}\n")
            summary = browse_website(url=res['href'], question=question)
            summaries.extend(summary)
            print(f"{res['title']} x {res['href']} summary: {summary}\n")
    else:
        print(f"{search_search_json['title']} x {search_search_json['href']}")
        summary = browse_website(url=search_search_json['href'], question=question)
        summaries.extend(summary)
        print(f"{search_search_json['title']} x {search_search_json['href']} summary: {summary}\n")

    if len(summaries) == 1:
        return summaries[0]

    if len(summaries) == 1:
        return summaries[0]

    combined_summary = "\n".join(summaries)
    messages = [create_message(combined_summary, question)]

    return create_chat_completion(
        model=CFG.fast_llm_model,
        messages=messages,
    )